/* slhttp - simple/limited HTTP client to GET specified URI
   Copyright (C) 2004 kana <nicht AT s8 DOT xrea DOT con>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if SLHTTP_USE_WINSOCK
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#endif

#if !(SLHTTP_USE_WINSOCK)
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
typedef int SOCKET;
#define INVALID_SOCKET (-1)
#define closesocket close
#endif

#include "slhttp.h"




#define NUMBER_OF(array) (sizeof(array) / sizeof((array)[0]))




/**
 * Split `http_uri' into `host', `port' and `path'.
 * Return SLHTTP_E_SUCCESS (= 0) for success or other value for failure.
 *
 * `host' and `path' will be stored a string allocated by malloc(),
 * so you MUST free() if they are no longer used.
 *
 * Regardless of the result,
 * `host', `port' and `path' will be modified.
 */

static int
split_http_uri( const char* http_uri,
                char** host, unsigned short* port, char** path )
{
	char* slash_slash;
	char* slash;
	int host_len;
	char* port_s;
	int result;

	if ( (strncmp(http_uri, "http://", 7) != 0)
	     && (strncmp(http_uri, "https://", 8) != 0) )
	{
		result = SLHTTP_E_INVALID_SCHEME;
		goto e_scheme;
	}

	/**
	 * slash_slash                  slash
	 *      |                         |
	 *      v                         v
	 * http://gekimoe.mahousyoujo.info/2004/10/15/21/49/ctime
	 */

	slash_slash = strstr(http_uri, "//");
	slash = strstr(slash_slash + 2, "/");
	if (slash == NULL) {
		result = SLHTTP_E_INVALID_URI;
		goto e_path;
	}

	host_len = (slash - (slash_slash + 2)) / sizeof(char);
	*host = malloc(sizeof(char) * (host_len + 1));
	if (*host == NULL) {
		result = SLHTTP_E_NO_MEMORY;
		goto e_malloc_host;
	}
	strncpy(*host, slash_slash + 2, host_len);
	(*host)[host_len] = '\0';

	*path = malloc(sizeof(char) * (strlen(slash) + 1));
	if (*path == NULL) {
		result = SLHTTP_E_NO_MEMORY;
		goto e_malloc_path;
	}
	strcpy(*path, slash);

	port_s = strchr(*host, ':');
	if (port_s != NULL) {  /* host includes port number */
		port_s[0] = '\0';
		port_s++;
		*port = atoi(port_s);
	} else {
		*port = 80;
	}

	return SLHTTP_E_SUCCESS;

	free(*path);
e_malloc_path:
	free(*host);
e_malloc_host:
e_path:
e_scheme:
	return result;
}




int
slhttp_get( const char* http_uri, const char* user_agent,
            int* status_code, char** status_line,
            char** headers,
            char** message_body )
{
	char* host;
	unsigned short port;
	char* path;

	struct hostent* hostent;
	struct sockaddr_in server;
	SOCKET sockd;
	char buf[1024];
	char* whole_response;
	int len;

	int result;

	result = split_http_uri(http_uri, &host, &port, &path);
	if (result != SLHTTP_E_SUCCESS) {
		/* result = SLHTTP_E_XXX;  -- already set by split_http_uri() */
		goto e_split_http_uri;
	}

	hostent = gethostbyname(host);
	if (hostent == NULL) {
		result = SLHTTP_E_GETHOSTBYNAME;
		goto e_gethostbyname;
	}

	memset(&server, 0x00, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	memcpy(&(server.sin_addr), hostent->h_addr, hostent->h_length);

	sockd = socket(PF_INET, SOCK_STREAM, 0);
	if (sockd == INVALID_SOCKET) {
		result = SLHTTP_E_SUCCESS;
		goto e_socket;
	}

	if (connect(sockd, (struct sockaddr*)(&server), sizeof(server)) != 0) {
		result = SLHTTP_E_CONNECT;
		goto e_connect;
	}

	/* Send request */
	snprintf( buf, NUMBER_OF(buf),
	          "GET %s HTTP/1.0\r\n"
	          "Host: %s:%d\r\n"
	          "User-Agent: %s\r\n"
	          "\r\n",
	          path,
	          host, port,
	          user_agent );
	send(sockd, buf, strlen(buf) / sizeof(char), 0);

	/* Receive response */
	whole_response = NULL;
	len = 0;
	while (!0) {
		char* t;
		int l;

		l = recv(sockd, buf, strlen(buf) / sizeof(char), 0);
		if (l == 0)
			break;

		t = realloc(whole_response, sizeof(char) * (len + l + 1));
		if (t == NULL) {
			result = SLHTTP_E_NO_MEMORY;
			goto e_receive_nomemory;
		}

		whole_response = t;
		memcpy(whole_response + len, buf, l);
		len += l;
	}
	if (whole_response == NULL) {
		result = SLHTTP_E_INVALID_HTTP_RESPONSE;
		goto e_parse_invalid;
	}
	whole_response[len] = '\0';

	/* Parse response*/

	{
		/**
		 * HTTP/1.1 200 OK\r\nHeaders: xxx\r\n\r\nbody...
		 * ^              ^               ^
		 * |              |               |
		 * whole_response first_crlf first_crlfcrlf
		 * |              |  _____________/
		 * v              v/'
		 * HTTP/1.1 200 OK\r\n\r\n(with no header)
		 */
		char* first_crlf;
		char* first_crlfcrlf;

		first_crlf = strstr(whole_response, "\r\n");
		first_crlfcrlf = strstr(whole_response, "\r\n\r\n");

		if (first_crlf == NULL || first_crlfcrlf == NULL) {
			result = SLHTTP_E_INVALID_HTTP_RESPONSE;
			goto e_parse_invalid;
		}

		if (1 != sscanf( whole_response,
		                 "HTTP/%*d.%*d %3d %*s\r\n",
		                 status_code ))
		{
			result = SLHTTP_E_INVALID_HTTP_RESPONSE;
			goto e_parse_invalid;
		}

		first_crlf[0] = '\0';
		first_crlf[1] = '\0';
		first_crlfcrlf[0] = '\0';
		first_crlfcrlf[1] = '\0';
		first_crlfcrlf[2] = '\0';
		first_crlfcrlf[3] = '\0';

		*status_line = strdup(whole_response);
		*headers = strdup(first_crlf + 2);
		*message_body = strdup(first_crlfcrlf + 4);
		if ( *status_line == NULL || *headers == NULL
		     || *message_body == NULL )
		{
			result = SLHTTP_E_NO_MEMORY;
			free(*status_line);
			free(*headers);
			free(*message_body);
			goto e_parse_nomemory;
		}
	}

	result = SLHTTP_E_SUCCESS;

e_parse_nomemory:
e_parse_invalid:
	free(whole_response);
e_receive_nomemory:
e_connect:
	closesocket(sockd);
e_socket:
e_gethostbyname:
	free(host);
	free(path);
e_split_http_uri:
	return result;
}




#ifdef __TEST__

int
main(int ac, char** av)
{
	int status_code;
	char* status_line;
	char* headers;
	char* message_body;
	int e;
#if SLHTTP_USE_WINSOCK
	WSADATA dummy;
#endif

	if (ac - 1 != 1) {
		puts("Usage: slhttp http-uri");
		return EXIT_FAILURE;
	}

#if SLHTTP_USE_WINSOCK
	if (WSAStartup(MAKEWORD(2, 2), &dummy) != 0) {
		puts("WSAStartUp failed");
		return EXIT_FAILURE;
	}
#endif

	e = slhttp_get( av[1], "slhttp/0.0.0",
	                &status_code, &status_line,
	                &headers, &message_body );
	if (e == SLHTTP_E_SUCCESS) {
		printf( "Status line\n"
		        "-----------\n"
		        "%s\n"
		        "\n",
		        status_line );
		printf( "Headers\n"
		        "-------\n"
		        "%s\n"
		        "\n",
		        headers );
		printf( "Message body\n"
		        "------------\n"
		        "%s"
		        "\n",
		        message_body );
		free(status_line);
		free(headers);
		free(message_body);
	} else {
		perror("slhttp_get()");
	}

#if SLHTTP_USE_WINSOCK
	WSACleanup();
#endif


	return EXIT_SUCCESS;
}

#endif




/* __END__ */
